# -*- coding: utf-8 -*-
import falcon
from graceful.parameters import IntParam
from graceful.resources.base import BaseResource


class BaseMixin():
    """Base mixin class."""

    def handle(self, handler, req, resp, **kwargs):
        """Handle given resource manipulation flow in consistent manner.

        This mixin is intended to be used only as a base class in new flow
        mixin classes. It ensures that regardless of resource manunipulation
        semantics (retrieve, get, delete etc.) the flow is always the same:

        1. Decode and validate all request parameters from the query string
           using ``self.require_params()`` method.
        2. Use ``self.require_meta_and_content()`` method to construct ``meta``
           and ``content`` dictionaries that will be later used to create
           serialized response body.
        3. Construct serialized response body using ``self.body()`` method.

        Args:
             handler (method): resource manipulation method handler.
             req (falcon.Request): request object instance.
             resp (falcon.Response): response object instance to be modified.
             **kwargs: additional keyword arguments retrieved from url
                 template.

        Returns:
             Content dictionary (preferably resource representation).
        """
        params = self.require_params(req)

        meta, content = self.require_meta_and_content(
            handler, params, **kwargs
        )
        self.make_body(resp, params, meta, content)
        return content


class RetrieveMixin(BaseMixin):
    """Add default "retrieve flow on GET" to any resource class."""

    def retrieve(self, params, meta, **kwargs):
        """Retrieve existing resource instance and return its representation.

        Value returned by this handler will be included in response
        'content' section.

        Args:
            params (dict): dictionary of parsed parameters accordingly
                to definitions provided as resource class atributes.
            meta (dict): dictionary of meta parameters anything added
                to this dict will will be later included in response
                'meta' section. This can already prepopulated by method
                that calls this handler.
            **kwargs: dictionary of values retrieved from route url
                template by falcon. This is suggested way for providing
                resource identifiers.


        Returns:
            value to be included in response 'content' section

        """
        raise NotImplementedError("retrieve method not implemented")

    def on_get(self, req, resp, handler=None, **kwargs):
        """Respond on GET HTTP request assuming resource retrieval flow.

        This request handler assumes that GET requests are associated with
        single resource instance retrieval. Thus default flow for such requests
        is:

        * Retrieve single resource instance of prepare its representation by
          calling retrieve method handler.

        Args:
            req (falcon.Request): request object instance.
            resp (falcon.Response): response object instance to be modified
            handler (method): list method handler to be called. Defaults
                to ``self.list``.
            **kwargs: additional keyword arguments retrieved from url template.
        """
        self.handle(
            handler or self.retrieve, req, resp, **kwargs
        )


class ListMixin(BaseMixin):
    """Add default "list flow on GET" to any resource class."""

    def list(self, params, meta, **kwargs):
        """List existing resource instances and return their representations.

        Value returned by this handler will be included in response
        'content' section.

        Args:
            params (dict): dictionary of parsed parameters accordingly
               to definitions provided as resource class atributes.
            meta (dict): dictionary of meta parameters anything added
               to this dict will will be later included in response
               'meta' section. This can already prepopulated by method
               that calls this handler.
            **kwargs: dictionary of values retrieved from route url
               template by falcon. This is suggested way for providing
               resource identifiers.


        Returns:
            value to be included in response 'content' section

        """
        raise NotImplementedError("list method not implemented")

    def on_get(self, req, resp, handler=None, **kwargs):
        """Respond on GET HTTP request assuming resource list retrieval flow.

        This request handler assumes that GET requests are associated with
        resource list retrieval. Thus default flow for such requests is:

        * Retrieve list of existing resource instances and prepare their
          representations by calling list retrieval method handler.

        Args:
            req (falcon.Request): request object instance.
            resp (falcon.Response): response object instance to be modified
            handler (method): list method handler to be called. Defaults
                to ``self.list``.
            **kwargs: additional keyword arguments retrieved from url template.
        """
        self.handle(
            handler or self.list, req, resp, **kwargs
        )


class DeleteMixin(BaseMixin):
    """Add default "delete flow on DELETE" to any resource class."""

    def delete(self, params, meta, **kwargs):
        """Delete existing resource instance.

        Args:
            params (dict): dictionary of parsed parameters accordingly
               to definitions provided as resource class atributes.
            meta (dict): dictionary of meta parameters anything added
               to this dict will will be later included in response
               'meta' section. This can already prepopulated by method
               that calls this handler.
            **kwargs: dictionary of values retrieved from route url
               template by falcon. This is suggested way for providing
               resource identifiers.

        Returns:
            value to be included in response 'content' section

        """
        raise NotImplementedError("delete method not implemented")

    def on_delete(self, req, resp, handler=None, **kwargs):
        """Respond on DELETE HTTP request assuming resource deletion flow.

        This request handler assumes that DELETE requests are associated with
        resource deletion. Thus default flow for such requests is:

        * Delete existing resource instance.
        * Set response status code to ``202 Accepted``.

        Args:
            req (falcon.Request): request object instance.
            resp (falcon.Response): response object instance to be modified
            handler (method): deletion method handler to be called. Defaults
                to ``self.delete``.
            **kwargs: additional keyword arguments retrieved from url template.
        """
        self.handle(
            handler or self.delete, req, resp, **kwargs
        )

        resp.status = falcon.HTTP_ACCEPTED


class UpdateMixin(BaseMixin):
    """Add default "update flow on PUT" to any resource class."""

    def update(self, params, meta, **kwargs):
        """Update existing resource instance and return its representation.

        Value returned by this handler will be included in response
        'content' section.

        Args:
            params (dict): dictionary of parsed parameters accordingly
                to definitions provided as resource class atributes.
            meta (dict): dictionary of meta parameters anything added
                to this dict will will be later included in response
                'meta' section. This can already prepopulated by method
                that calls this handler.
            **kwargs: dictionary of values retrieved from route url
                template by falcon. This is suggested way for providing
                resource identifiers.

        Returns:
            value to be included in response 'content' section

        """
        raise NotImplementedError("update method not implemented")

    def on_put(self, req, resp, handler=None, **kwargs):
        """Respond on PUT HTTP request assuming resource update flow.

        This request handler assumes that PUT requests are associated with
        resource update/modification. Thus default flow for such requests is:

        * Modify existing resource instance and prepare its representation by
          calling its update method handler.
        * Set response status code to ``202 Accepted``.

        Args:
            req (falcon.Request): request object instance.
            resp (falcon.Response): response object instance to be modified
            handler (method): update method handler to be called. Defaults
                to ``self.update``.
            **kwargs: additional keyword arguments retrieved from url template.
        """
        self.handle(
            handler or self.update, req, resp, **kwargs
        )
        resp.status = falcon.HTTP_ACCEPTED


class CreateMixin(BaseMixin):
    """Add default "creation flow on POST" to any resource class."""

    def create(self, params, meta, **kwargs):
        """Create new resource instance and return its representation.

        This is default resource instance creation method. Value returned
        is the representation of single resource instance. It will be included
        in the 'content' section of response body.

        Args:
            params (dict): dictionary of parsed parameters accordingly
                to definitions provided as resource class atributes.
            meta (dict): dictionary of meta parameters anything added
                to this dict will will be later included in response
                'meta' section. This can already prepopulated by method
                that calls this handler.
            kwargs (dict): dictionary of values retrieved from route url
                template by falcon. This is suggested way for providing
                resource identifiers.


        Returns:
            value to be included in response 'content' section

        """
        raise NotImplementedError("create method not implemented")

    def get_object_location(self, obj):
        """Return location URI associated with given resource representation.

        This handler is optional. Returned URI will be included as the
        value of ``Location`` header on POST responses.
        """
        raise NotImplementedError("update method not implemented")

    def on_post(self, req, resp, handler=None, **kwargs):
        """Respond on POST HTTP request assuming resource creation flow.

        This request handler assumes that POST requests are associated with
        resource creation. Thus default flow for such requests is:

        * Create new resource instance and prepare its representation by
          calling its creation method handler.
        * Try to retrieve URI of newly created object using
          ``self.get_object_location()``. If it succeeds use that URI as the
          value of ``Location`` header in response object instance.
        * Set response status code to ``201 Created``.

        Args:
            req (falcon.Request): request object instance.
            resp (falcon.Response): response object instance to be modified
            handler (method): creation method handler to be called. Defaults
                to ``self.create``.
            **kwargs: additional keyword arguments retrieved from url template.
        """
        obj = self.handle(
            handler or self.create, req, resp, **kwargs
        )
        try:
            resp.location = self.get_object_location(obj)
        except NotImplementedError:
            pass

        resp.status = falcon.HTTP_CREATED


class PaginatedMixin(BaseResource):
    """Add simple pagination capabilities to resource.

    This class provides two additional parameters with some default
    descriptions and ``add_pagination_meta`` method that can update
    meta with more useful pagination information.

    Example usage:

    .. code-block:: python

        from graceful.resources.mixins import PaginatedMixin
        from graceful.resources.generic import ListResource

        class SomeResource(PaginatedMixin, ListResource):

            def list(self, params, meta):
                # params has now 'page' and 'page_size' params that
                # can be used for offset&limit-like operations
                self.add_pagination_meta(params, meta)

                # ...

    """

    page_size = IntParam(
        details="""Specifies number of result entries in single response""",
        default='10'
    )
    page = IntParam(
        details="""Specifies number of results page for response.
        Page count starts from 0""",
        default='0',
    )

    def add_pagination_meta(self, params, meta):
        """Extend default meta dictionary value with pagination hints.

        Note:
            This method handler attaches values to ``meta`` dictionary without
            changing it's reference. This means that you should never replace
            ``meta`` dictionary with any other dict instance but simply modify
            its content.

        Args:
            params (dict): dictionary of decoded parameter values
            meta (dict): dictionary of meta values attached to response
        """
        meta['page_size'] = params['page_size']
        meta['page'] = params['page']

        meta['prev'] = "page={0}&page_size={1}".format(
            params['page'] - 1, params['page_size']
        ) if meta['page'] > 0 else None

        meta['next'] = "page={0}&page_size={1}".format(
            params['page'] + 1, params['page_size']
        ) if meta.get('has_more', True) else None



# -*- coding: utf-8 -*-
from functools import partial

from graceful.resources.base import BaseResource
from graceful.resources.mixins import (
    RetrieveMixin,
    ListMixin,
    UpdateMixin,
    CreateMixin,
    DeleteMixin,
    PaginatedMixin,
)


class Resource(RetrieveMixin, BaseResource):
    """Basic retrieval of resource instance lists without serialization.

    This resource class is intended for endpoints that do not require automatic
    representation serialization and extensive field descriptions but still
    gives support for defining parameters as resource class attributes.

    Example usage:

    .. code-block:

        from graceful.resources.generic import Resource
        from graceful.parameters import StringParam

        class SampleResource(Resource):
            dummy = StringParam("some example dummy parameter")

            def retrieve(self, params, meta):
                return {"sample": "resource"}

    """


class ListResource(ListMixin, BaseResource):
    """Basic retrieval of resource instance lists without serialization.

    This resource class is intended for endpoints that do not require automatic
    representation serialization and extensive field descriptions but still
    gives support for defining parameters as resource class attributes.

    Example usage:

    .. code-block:

        from graceful.resources.generic import ListResource
        from graceful.parameters import StringParam

        class SampleResource(ListResource):
            some_filter = StringParam("some example filter parameter")

            def list(self, params, meta):
                return [{"sample": "resource"}]

    """


class RetrieveAPI(RetrieveMixin, BaseResource):
    """Generic Retrieve API with resource serialization.

    Generic resource that uses serializer for resource description,
    serialization and validation.

    Allowed methods:

    * GET: retrieve resource representation (handled with ``.retrieve()``
      method handler)

    """

    serializer = None

    def describe(self, req=None, resp=None, **kwargs):
        """Extend default endpoint description with serializer description."""
        return super().describe(
            req, resp,
            type='object',
            fields=self.serializer.describe() if self.serializer else None,
            **kwargs
        )

    def _retrieve(self, params, meta, **kwargs):
        return self.serializer.to_representation(
            self.retrieve(params, meta, **kwargs)
        )

    def on_get(self, req, resp, **kwargs):
        """Respond on GET requests using ``self.retrieve()`` handler."""
        return super().on_get(
            req, resp, handler=self._retrieve, **kwargs
        )


class RetrieveUpdateAPI(UpdateMixin, RetrieveAPI):
    """Generic Retrieve/Update API with resource serialization.

    Generic resource that uses serializer for resource description,
    serialization and validation.

    Allowed methods:

    * GET: retrieve resource representation handled with ``.retrieve()``
      method handler
    * PUT: update resource with representation provided in request body
      (handled with ``.update()`` method handler)

    """

    def _update(self, params, meta, **kwargs):
        return self.serializer.to_representation(
            self.update(params, meta, **kwargs)
        )

    def on_put(self, req, resp, **kwargs):
        """Respond on PUT requests using ``self.update()`` handler."""
        validated = self.require_validated(req)
        return super().on_put(
            req, resp,
            handler=partial(self._update, validated=validated),
            **kwargs
        )


class RetrieveUpdateDeleteAPI(DeleteMixin, RetrieveUpdateAPI):
    """Generic Retrieve/Update/Delete API with resource serialization.

    Generic resource that uses serializer for resource description,
    serialization and validation.

    Allowed methods:

    * GET: retrieve resource representation (handled with ``.retrieve()``
      method handler)
    * PUT: update resource with representation provided in request body
      (handled with ``.update()`` method handler)
    * DELETE: delete resource (handled with ``.delete()`` method handler)

    """


class ListAPI(ListMixin, BaseResource):
    """Generic List API with resource serialization.

    Generic resource that uses serializer for resource description,
    serialization and validation.

    Allowed methods:

    * GET: list multiple resource instances representations (handled
      with ``.list()`` method handler)

    """

    def _list(self, params, meta, **kwargs):
        return [
            self.serializer.to_representation(obj)
            for obj in self.list(params, meta, **kwargs)
        ]

    def describe(self, req=None, resp=None, **kwargs):
        """Extend default endpoint description with serializer description."""
        return super().describe(
            req, resp,
            type='list',
            fields=self.serializer.describe() if self.serializer else None,
            **kwargs
        )

    def on_get(self, req, resp, **kwargs):
        """Respond on GET requests using ``self.list()`` handler."""
        return super().on_get(req, resp, handler=self._list, **kwargs)


class ListCreateAPI(CreateMixin, ListAPI):
    """Generic List/Create API with resource serialization.

    Generic resource that uses serializer for resource description,
    serialization and validation.

    Allowed methods:

    * GET: list multiple resource instances representations (handled
      with ``.list()`` method handler)
    * POST: create new resource from representation provided in request body
      (handled with ``.create()`` method handler)

    """

    def _create(self, params, meta, **kwargs):
        return self.serializer.to_representation(
            self.create(params, meta, **kwargs)
        )

    def on_post(self, req, resp, **kwargs):
        """Respond on POST requests using ``self.create()`` handler."""
        validated = self.require_validated(req)

        return super().on_post(
            req, resp,
            handler=partial(self._create, validated=validated),
            **kwargs
        )


class PaginatedListAPI(PaginatedMixin, ListAPI):
    """Generic List API with resource serialization and pagination.

    Generic resource that uses serializer for resource description,
    serialization and validation.

    Adds simple pagination to list of resources.

    Allowed methods:

    * GET: list multiple resource instances representations (handled
      with ``.list()`` method handler)

    """

    def _list(self, params, meta, **kwargs):
        objects = super()._list(params, meta, **kwargs)
        # note: we need to populate meta after objects are retrieved
        self.add_pagination_meta(params, meta)
        return objects


class PaginatedListCreateAPI(PaginatedMixin, ListCreateAPI):
    """Generic List/Create API with resource serialization and pagination.

    Generic resource that uses serializer for resource description,
    serialization and validation.

    Adds simple pagination to list of resources.

    Allowed methods:

    * GET: list multiple resource instances representations (handled
      with ``.list()`` method handler)
    * POST: create new resource from representation provided in request body
      (handled with ``.create()`` method handler)

    """

    def _list(self, params, meta, **kwargs):
        objects = super()._list(
            params, meta, **kwargs
        )
        # note: we need to populate meta after objects are retrieved
        self.add_pagination_meta(params, meta)
        return objects



# -*- coding: utf-8 -*-
import base64
import decimal
import binascii
import inspect


class BaseParam():
    """Base parameter class for subclassing.

    To create new parameter type subclass ``BaseParam`` and implement
    ``.value()`` method handler.

    Args:
        details (str): verbose description of parameter. Should contain all
            information that may be important to your API user and will be used
            for describing resource on ``OPTIONS`` requests and ``.describe()``
            call.

        label (str): human readable label for this parameter (it will be used
            for describing resource on OPTIONS requests).

            *Note that it is recomended to use parameter names that are
            self-explanatory intead of relying on param labels.*

        required (bool): if set to ``True`` then all GET, POST, PUT,
            PATCH and DELETE requests will return ``400 Bad Request`` response
            if query param is not provided. Defaults to ``False``.

        default (str): set default value for param if it is not
            provided in request as query parameter. This MUST be a raw string
            value that will be then parsed by ``.value()`` handler.

            If default is set and ``required`` is ``True`` it will raise
            ``ValueError`` as having required parameters with default
            value has no sense.

        many (str): set to ``True`` if multiple occurences of this parameter
            can be included in query string, as a result values for this
            parameter will be always included as a list in params dict.
            Defaults to ``False``.

          .. note::
              If ``many=False`` and client inlcudes multiple values for this
              parameter in query string then only one of those values will be
              returned, and it is undefined which one.

    Example:

    .. code-block:: python

           class BoolParam(BaseParam):
               def value(self, data):
                   if data in {'true', 'True', 'yes', '1', 'Y'}:
                       return True
                   elif data in {'false', 'False', 'no', '0', 'N'}:
                       return False
                   else:
                       raise ValueError(
                           "{data} is not valid boolean field".format(
                               data=data
                           )
                       )

    """

    spec = None
    type = None

    container = list

    def __init__(
            self,
            details,
            label=None,
            required=False,
            default=None,
            many=False,
            validators=None
    ):
        """Initialize parameter and verify default/required contraints."""
        self.label = label
        self.details = details
        self.required = required
        self.many = many
        self.validators = validators or []

        if not (default is None) and not isinstance(default, str):
            raise TypeError(
                "value for {cls} 'default' argument must be string instance"
                "".format(cls=self.__class__.__name__)
            )

        if not (default is None) and required:
            raise ValueError(
                "{cls}(required={required}, default='{default}'): "
                "initialization with both required and default makes no sense"
                "".format(
                    cls=self.__class__.__name__,
                    default=default,
                    required=required,
                )
            )
        self.default = default

    def validated_value(self, raw_value):
        """Return parsed parameter value and run validation handlers.

        Error message included in exception will be included in http error
        response

        Args:
            value: raw parameter value to parse validate

        Returns:
            None

        Note:
            Concept of validation for params is understood here as a process
            of checking if data of valid type (successfully parsed/processed by
            ``.value()`` handler) does meet some other constraints
            (lenght, bounds, uniqueness, etc.). It will internally call its
            ``value()`` handler.

        """
        value = self.value(raw_value)
        try:
            for validator in self.validators:
                validator(value)
        except:
            raise
        else:
            return value

    def value(self, raw_value):
        """Raw value deserialization method handler.

        Args:
            raw_value (str): raw value from GET parameters

        """
        raise NotImplementedError(
            "{cls}.value() method not implemented".format(
                cls=self.__class__.__name__
            )
        )

    def describe(self, **kwargs):
        """Describe this parameter instance for purpose of self-documentation.

        Args:
            kwargs (dict): dictionary of additional description items for
                extending default description

        Returns:
            dict: dictionary of description items


        Suggested way for overriding description fields or extending it with
        additional items is calling super class method with new/overriden
        fields passed as keyword arguments like following:

        .. code-block:: python

            class DummyParam(BaseParam):
                def description(self, **kwargs):
                    super().describe(is_dummy=True, **kwargs)

        """
        description = {
            'label': self.label,
            # note: details are expected to be large so it should
            #       be reformatted
            'details': inspect.cleandoc(self.details),
            'required': self.required,
            'many': self.many,
            'spec': self.spec,
            'default': self.default,
            'type': self.type or 'unspecified'
        }

        description.update(kwargs)
        return description


class StringParam(BaseParam):
    r"""Describes parameter that will always be returned as-is (string).

     Additional validation can be added to param instance using ``validators``
     argument during initialization:

    .. code-block:: python

        from graceful.parameters import StringParam
        from graceful.validators import match_validator
        from graceful.resources.generic import Resource

        class ExampleResource(Resource):
            word = StringParam(
                'one "word" parameter',
                validators=[match_validator('\w+')],
            )

    """

    type = 'string'

    def value(self, raw_value):
        """Return param value as-is (str)."""
        return raw_value


class Base64EncodedParam(BaseParam):
    """Describes string parameter with value encoded using Base64 encoding."""

    spec = (
        "RFC-4648 Section 4",
        "https://tools.ietf.org/html/rfc4648#section-4",
    )

    def value(self, raw_value):
        """Decode param with Base64."""
        try:
            return base64.b64decode(bytes(raw_value, 'utf-8')).decode('utf-8')
        except binascii.Error as err:
            raise ValueError(str(err))


class IntParam(BaseParam):
    """Describes parameter with value expressed as integer number."""

    type = "integer"

    def value(self, raw_value):
        """Decode param as integer value."""
        return int(raw_value)


class FloatParam(BaseParam):
    """Describes parameter with value expressed as float number."""

    type = "float"

    def value(self, raw_value):
        """Decode param as float value."""
        return float(raw_value)


class DecimalParam(BaseParam):
    """Describes parameter with value expressed as decimal number."""

    type = "decimal"

    def value(self, raw_value):
        """Decode param as decimal value."""
        try:
            return decimal.Decimal(raw_value)
        except decimal.InvalidOperation:
            raise ValueError(
                "Could not parse '{}' value as decimal".format(raw_value)
            )


class BoolParam(BaseParam):
    """Describes parameter with value expressed as bool.

    Accepted string values for boolean parameters are as follows:

    * False: ``['True', 'true', 'TRUE', 'T', 't', '1'}``
    * True: ``['False', 'false', 'FALSE', 'F', 'f', '0', '0.0']``

    In case raw parameter value does not match any of these strings the
    ``value()`` method will raise ``ValueError`` method.

    """

    type = "bool"

    _TRUE_VALUES = {'True', 'true', 'TRUE', 'T', 't', '1'}
    _FALSE_VALUES = {'False', 'false', 'FALSE', 'F', 'f', '0', '0.0'}

    def value(self, raw_value):
        """Decode param as bool value."""
        if raw_value in self._FALSE_VALUES:
            return False
        elif raw_value in self._TRUE_VALUES:
            return True
        else:
            raise ValueError(
                "Could not parse '{}' value as boolean".format(raw_value)
            )



# -*- coding: utf-8 -*-
import json
import inspect
from collections import OrderedDict

from falcon import errors
import falcon
from mimeparse import parse_mime_type

from graceful.parameters import BaseParam, IntParam
from graceful.errors import DeserializationError, ValidationError


class MetaResource(type):
    """Metaclass for handling parametrization with parameter objects."""

    _params_storage_key = '_params'

    @classmethod
    def __prepare__(mcs, name, bases, **kwargs):
        """Prepare class namespace in a way that ensures order of attributes.

        This needs to be an `OrderedDict` so `_get_params()` method can
        construct params storage that preserves the same order of parameters
        as defined in code.

        Note: this is python3 thing and support for ordering of params in
        descriptions will not be backported to python2 even if this framework
        will get python2 support.

        Args:
            bases: all base classes of created resource class
            namespace (dict): namespace as dictionary of attributes

        """
        return OrderedDict()

    @classmethod
    def _get_params(mcs, bases, namespace):
        """Create params dictionary to be used in resource class namespace.

        Pop all parameter objects from attributes dict (namespace)
        and store them under _params_storage_key atrribute.
        Also collect all params from base classes in order that ensures
        params can be overriden.

        Args:
            bases: all base classes of created resource class
            namespace (dict): namespace as dictionary of attributes

        """
        params = [
            (name, namespace.pop(name))
            for name, attribute
            in list(namespace.items())
            if isinstance(attribute, BaseParam)
        ]

        for base in reversed(bases):
            if hasattr(base, mcs._params_storage_key):
                params = list(
                    getattr(base, mcs._params_storage_key).items()
                ) + params

        return OrderedDict(params)

    def __new__(mcs, name, bases, namespace):
        """Create new class object instance and alter its namespace."""
        namespace[mcs._params_storage_key] = mcs._get_params(bases, namespace)

        return super().__new__(
            # note: there is no need preserve order in namespace anymore so
            # we convert it explicitely to dict
            mcs, name, bases, dict(namespace)
        )


class BaseResource(metaclass=MetaResource):
    """Base resouce class with core param and response functionality.

    This base class handles resource responses, parameter deserialization,
    and validation of request included representations if serializer is
    defined.

    """

    indent = IntParam(
        """
        JSON output indentation. Set to 0 if output should not be formated.
        """,
        default='0'
    )

    serializer = None

    @property
    def params(self):
        """Return dictionary of parameter definition objects."""
        return getattr(self, self.__class__._params_storage_key)

    def make_body(self, resp, params, meta, content):
        """Construct response body in ``resp`` object using JSON serialization.

        Args:
            resp (falcon.Response): response object where to include
                serialized body
            params (dict): dictionary of parsed parameters
            meta (dict): dictionary of metadata to be included in 'meta'
                section of response
            content (dict): dictionary of response content (resource
                representation) to be included in 'content' section of response

        Returns:
            None

        """
        response = {
            'meta': meta,
            'content': content
        }
        resp.content_type = 'application/json'
        resp.body = json.dumps(
            response,
            indent=params['indent'] or None if 'indent' in params else None
        )

    def allowed_methods(self):
        """Return list of allowed HTTP methods on this resource.

        This is only for purpose of making resource description.

        Returns:
            list: list of allowed HTTP method names (uppercase)

        """
        return [
            method
            for method, allowed in (
                ('GET', hasattr(self, 'on_get')),
                ('POST', hasattr(self, 'on_post')),
                ('PUT', hasattr(self, 'on_put')),
                ('DELETE', hasattr(self, 'on_delete')),
                ('HEAD', hasattr(self, 'on_head')),
                ('OPTIONS', hasattr(self, 'on_options')),
            ) if allowed
        ]

    def describe(self, req=None, resp=None, **kwargs):
        """Describe API resource using resource introspection.

        Additional description on derrived resource class can be added using
        keyword arguments and calling ``super().decribe()`` method call
        like following:

        .. code-block:: python

             class SomeResource(BaseResource):
                 def describe(req, resp, **kwargs):
                     return super().describe(
                         req, resp, type='list', **kwargs
                      )

        Args:
            req (falcon.Request): request object
            resp (falcon.Response): response object
            kwargs (dict): dictionary of values created from resource url
                template

        Returns:
            dict: dictionary with resource descritpion information

        """
        description = {
            'params': OrderedDict([
                (name, param.describe())
                for name, param in self.params.items()
            ]),
            'details':
                inspect.cleandoc(
                    self.__class__.__doc__ or
                    "This resource does not have description yet"
                ),
            'name': self.__class__.__name__,
            'methods': self.allowed_methods()
        }
        # note: add path to resource description only if request object was
        #       provided in order to make auto-documentation engines simpler
        if req:
            description['path'] = req.path

        description.update(**kwargs)
        return description

    def on_options(self, req, resp, **kwargs):
        """Respond with JSON formatted resource description on OPTIONS request.

        Args:
            req (falcon.Request): Optional request object. Defaults to None.
            resp (falcon.Response): Optional response object. Defaults to None.
            kwargs (dict): Dictionary of values created by falcon from
                resource uri template.

        Returns:
            None

        """
        resp.set_header('Allow', ', '.join(self.allowed_methods()))
        resp.body = json.dumps(self.describe(req, resp))
        resp.content_type = 'application/json'

    def require_params(self, req):
        """Require all defined parameters from request query string.

        Raises ``falcon.errors.HTTPMissingParam`` exception if any of required
        parameters is missing and ``falcon.errors.HTTPInvalidParam`` if any
        of parameters could not be understood (wrong format).

        Args:
            req (falcon.Request): request object

        """
        params = {}

        for name, param in self.params.items():
            if name not in req.params and param.required:
                # we could simply raise with this single param or use get_param
                # with required=True parameter but for client convenience
                # we prefer to list all missing params that are required
                missing = set(
                    p for p in self.params
                    if self.params[p].required
                ) - set(req.params.keys())

                raise errors.HTTPMissingParam(", ".join(missing))

            elif name in req.params or param.default:
                # Note: lack of key in req.params means it was not specified
                # so unless there is default value it will not be included in
                # output params dict.
                # This way we have explicit information that param was
                # not specified. Using None would not be as good because param
                # class can also return None from `.value()` method as a valid
                # translated value.
                try:
                    if param.many:
                        # params with "many" enabled need special care
                        values = req.get_param_as_list(
                            # note: falcon allows to pass value handler using
                            #       `transform` param so we do not need to
                            #       iterate through list manually
                            name, param.validated_value
                        ) or [
                            param.default and
                            param.validated_value(param.default)
                        ]
                        params[name] = param.container(values)
                    else:
                        # note that if many==False and query parameter
                        # occurs multiple times in qs then it is
                        # **unspecified** which one will be used. See:
                        # http://falcon.readthedocs.org/en/latest/api/request_and_response.html#falcon.Request.get_param  # noqa
                        params[name] = param.validated_value(
                            req.get_param(name, default=param.default)
                        )

                except ValidationError as err:
                    # ValidationError allows to easily translate itself to
                    # to falcon's HTTPInvalidParam (Bad Request HTTP response)
                    raise err.as_invalid_param(name)

                except ValueError as err:
                    # Other parsing issues are expected to raise ValueError
                    raise errors.HTTPInvalidParam(str(err), name)

        return params

    def require_meta_and_content(self, content_handler, params, **kwargs):
        """Require 'meta' and 'content' dictionaries using proper hander.

        Args:
            content_handler (callable): function that accepts
                ``params, meta, **kwargs`` argument and returns dictionary
                for ``content`` response section
            params (dict): dictionary of parsed resource parameters
            kwargs (dict): dictionary of values created from resource url
                template

        Returns:
            tuple (meta, content): two-tuple with dictionaries of ``meta`` and
                ``content`` response sections

        """
        meta = {
            'params': params
        }
        content = content_handler(params, meta, **kwargs)
        meta['params'] = params
        return meta, content

    def require_representation(self, req):
        """Require raw representation dictionary from falcon request object.

        This does not perform any field parsing or validation but only uses
        allowed content-encoding handler to decode content body.

        Note:
            Currently only JSON is allowed as content type.

        Args:
            req (falcon.Request): request object

        Returns:
            dict: raw dictionary of representation supplied in request body

        """
        try:
            type_, subtype, _ = parse_mime_type(req.content_type)
            content_type = '/'.join((type_, subtype))
        except:
            raise falcon.HTTPUnsupportedMediaType(
                description="Invalid Content-Type header: {}".format(
                    req.content_type
                )
            )

        if content_type == 'application/json':
            body = req.stream.read()
            return json.loads(body.decode('utf-8'))
        else:
            raise falcon.HTTPUnsupportedMediaType(
                description="only JSON supported, got: {}".format(content_type)
            )

    def require_validated(self, req, partial=False):
        """Require fully validated internal object dictionary.

        Internal object dictionary creation is based on content-decoded
        representation retrieved from request body. Internal object validation
        is performed using resource serializer.

        Args:
            req (falcon.Request): request object
            partial (bool): self to True if partially complete representation
                is accepted (e.g. for patching instead of full update). Missing
                fields in representation will be skiped.

        Returns:
            dict: dictionary of fields and values representing internal object.
                Each value is a result of ``field.from_representation`` call.

        """
        representation = self.require_representation(req)

        try:
            object_dict = self.serializer.from_representation(representation)
            self.serializer.validate(object_dict, partial)

        except DeserializationError as err:
            # when working on Resource we know that we can finally raise
            # bad request exceptions
            raise err.as_bad_request()

        except ValidationError as err:
            # ValidationError is a suggested way to validate whole resource
            # so we also are prepared to catch it
            raise err.as_bad_request()

        return object_dict



# -*- coding: utf-8 -*-
from collections import OrderedDict
from collections.abc import Mapping, MutableMapping

from graceful.errors import DeserializationError, ValidationError
from graceful.fields import BaseField


class MetaSerializer(type):
    """Metaclass for handling serialization with field objects."""

    _fields_storage_key = '_fields'

    @classmethod
    def __prepare__(mcs, name, bases, **kwargs):
        """Prepare class namespace in a way that ensures order of attributes.

        This needs to be an `OrderedDict` so `_get_fields()` method can
        construct fields storage that preserves the same order of fields as
        defined in code.

        Note: this is python3 thing and support for ordering of params in
        descriptions will not be backported to python2 even if this framework
        will get python2 support.

        """
        return OrderedDict()

    @classmethod
    def _get_fields(mcs, bases, namespace):
        """Create fields dictionary to be used in resource class namespace.

        Pop all field objects from attributes dict (namespace) and store them
        under _field_storage_key atrribute. Also collect all fields from base
        classes in order that ensures fields can be overriden.

        Args:
            bases: all base classes of created serializer class
            namespace (dict): namespace as dictionary of attributes

        """
        fields = [
            (name, namespace.pop(name))
            for name, attribute
            in list(namespace.items())
            if isinstance(attribute, BaseField)
        ]

        for base in reversed(bases):
            if hasattr(base, mcs._fields_storage_key):
                fields = list(
                    getattr(base, mcs._fields_storage_key).items()
                ) + fields

        return OrderedDict(fields)

    def __new__(mcs, name, bases, namespace):
        """Create new class object instance and alter its namespace."""
        namespace[mcs._fields_storage_key] = mcs._get_fields(bases, namespace)
        return super().__new__(
            # note: there is no need preserve order in namespace anymore so
            # we convert it explicitely to dict
            mcs, name, bases, dict(namespace)
        )


class BaseSerializer(metaclass=MetaSerializer):
    """Base serializer class for describing internal object serialization.

    Example:

    .. code-block:: python

        from graceful.serializers import BaseSerializer
        from graceful.fields import RawField, IntField, FloatField


        class CatSerializer(BaseSerializer):
            species = RawField("non normalized cat species")
            age = IntField("cat age in years")
            height = FloatField("cat height in cm")

    """

    @property
    def fields(self):
        """Return dictionary of field definition objects of this serializer."""
        return getattr(self, self.__class__._fields_storage_key)

    def to_representation(self, obj):
        """Convert given internal object instance into representation dict.

        Representation dict may be later serialized to the content-type
        of choice in the resource HTTP method handler.

        This loops over all fields and retrieves source keys/attributes as
        field values with respect to optional field sources and converts each
        one using ``field.to_representation()`` method.

        Args:
            obj (object): internal object that needs to be represented

        Returns:
            dict: representation dictionary

        """
        representation = {}

        for name, field in self.fields.items():
            # note fields do not know their names in source representation
            # but may know what attribute they target from source object
            attribute = self.get_attribute(obj, field.source or name)

            if attribute is None:
                # Skip none attributes so fields do not have to deal with them
                representation[name] = [] if field.many else None
            elif field.many:
                representation[name] = [
                    field.to_representation(item) for item in attribute
                ]
            else:
                representation[name] = field.to_representation(attribute)

        return representation

    def from_representation(self, representation):
        """Convert given representation dict into internal object.

        Internal object is simply a dictionary of values with respect to field
        sources.

        This does not check if all required fields exist or values are
        valid in terms of value validation
        (see: :meth:`BaseField.validate()`) but still requires all of passed
        representation values to be well formed representation (success call
        to ``field.from_representation``).

        In case of malformed representation it will run additional validation
        only to provide a full detailed exception about all that might be
        wrong with provided representation.

        Args:
           representation (dict): dictionary with field representation values

        Raises:
            DeserializationError: when at least one representation field
                is not formed as expected by field object. Information
                about additional forbidden/missing/invalid fields is provided
                as well.

        """
        object_dict = {}
        failed = {}

        for name, field in self.fields.items():
            if name not in representation:
                continue

            try:
                # if field has explicitely specified source then use it
                # else fallback to field name.
                # Note: field does not know its name
                object_dict[field.source or name] = field.from_representation(
                    representation[name]
                )
            except ValueError as err:
                failed[name] = str(err)

        if failed:
            # if failed to parse we eagerly perform validation so full
            # information about what is wrong will be returned
            try:
                self.validate(object_dict)
                # note: this exception can be reached with partial==True
                # since do not support partial updates yet this has 'no cover'
                raise DeserializationError()  # pragma: no cover
            except DeserializationError as err:
                err.failed = failed
                raise

        return object_dict

    def validate(self, object_dict, partial=False):
        """Validate given internal object returned by ``to_representation()``.

        Internal object is validated against missing/forbidden/invalid fields
        values using fields definitions defined in serializer.

        Args:
            object_dict (dict): internal object dictionart to perform
              to validate
            partial (bool): if set to True then incomplete object_dict
              is accepter and will not raise any exceptions when one
              of fields is missing

        Raises:
            DeserializationError:

        """
        # we are working on object_dict not an representation so there
        # is a need to annotate sources differently
        sources = {
            # TODO: handling of '*' sources here is a bit terryfying
            # TODO: maybe this needs more care in future releases
            field.source or name if field.source != "*" else name: field
            for name, field in self.fields.items()
        }

        # note: we are checking for all mising and invalid fields so we can
        # return exception with all fields that are missing and should
        # exist instead of single one
        missing = [
            name for name, field in sources.items()
            if all((not partial, name not in object_dict, not field.read_only))
        ]

        forbidden = [
            name for name in object_dict
            if any((name not in sources, sources[name].read_only))
        ]

        invalid = {}
        for name, value in object_dict.items():
            try:
                sources[name].validate(value)
            except ValidationError as err:
                invalid[name] = str(err)

        if any([missing, forbidden, invalid]):
            raise DeserializationError(missing, forbidden, invalid)

    def get_attribute(self, obj, attr):
        """Get attribute of given object instance.

        Reason for existence of this method is the fact that  'attribute' can
        be also object's key from if is a dict or any other kind of mapping.

        Note: it will return None if attribute key does not exist

        Args:
            obj (object): internal object to retrieve data from

        Returns:
            internal object's key value or attribute

        """
        # '*' is a special wildcard character that means whole object
        # is passed
        if attr == '*':
            return obj

        # if this is any mapping then instead of attributes use keys
        if isinstance(obj, Mapping):
            return obj.get(attr, None)

        return getattr(obj, attr, None)

    def set_attribute(self, obj, attr, value):
        """Set value of attribute in given object instance.

        Reason for existence of this method is the fact that 'attribute' can
        be also a object's key if it is a dict or any other kind of mapping.

        Args:
            obj (object): object instance to modify
            attr (str): attribute (or key) to change
            value: value to set

        """
        # if this is any mutable mapping then instead of attributes use keys
        if isinstance(obj, MutableMapping):
            obj[attr] = value
        else:
            setattr(obj, attr, value)

    def describe(self):
        """Describe all serialized fields.

        It returns dictionary of all fields description defined for this
        serializer using their own ``describe()`` methods with respect to order
        in which they are defined as class attributes.

        Returns:
            OrderedDict: serializer description

        """
        return OrderedDict([
            (name, field.describe())
            for name, field in self.fields.items()
        ])



# -*- coding: utf-8 -*-
from falcon import HTTPBadRequest, HTTPInvalidParam


class DeserializationError(ValueError):
    """Raised when error happened during deserialization of representation."""

    def __init__(
            self, missing=None, forbidden=None, invalid=None, failed=None
    ):
        """Initialize exception instance."""
        self.missing = missing
        self.forbidden = forbidden
        self.invalid = invalid
        self.failed = failed

    def as_bad_request(self):
        """Translate this error to falcon's HTTP specific error exception."""
        return HTTPBadRequest(
            title="Representation deserialization failed",
            description=self._get_description()
        )

    def _get_description(self):
        """Return human readable description error description.

        This description should explain everything that went wrong during
        deserialization.

        """
        return ", ".join([
            part for part in [
                "missing: {}".format(self.missing) if self.missing else "",
                (
                    "forbidden: {}".format(self.forbidden)
                    if self.forbidden else ""
                ),
                "invalid: {}:".format(self.invalid) if self.invalid else "",
                (
                    "failed to parse: {}".format(self.failed)
                    if self.failed else ""
                )
            ] if part
        ])


class ValidationError(ValueError):
    """Raised when validation error occured."""

    def as_bad_request(self):
        """Translate this error to falcon's HTTP specific error exception.

        Note:
            Exceptions returned by this method should be used to inform about
            resource validation failures. In case of param validation
            failures the ``as_invalid_param()`` method should be used.

        """
        return HTTPBadRequest(
            title="Validation failed",
            description=str(self)
        )

    def as_invalid_param(self, param_name):
        """Translate this error to falcon's HTTP specific error exception.

        Note:
            Exceptions returned by this method should be used to inform about
            param validation failures. In case of resource validation
            failures the ``as_bad_request()`` method should be used.

        Args:
            param_name (str): HTTP query string parameter name

        """
        return HTTPInvalidParam(
            str(self), param_name
        )



# -*- coding: utf-8 -*-
"""Minimalist framework for self-descriptive REST APIs build on top of falcon.

It is inspired by Django REST Framework package. Mostly by how object
serialization is done but more emphasis is put on API to be self-descriptive.
"""
VERSION = (0, 2, 0)  # PEP 386  # noqa
__version__ = ".".join([str(x) for x in VERSION])  # noqa
